// StarPlot - A program for interactively viewing 3D maps of stellar positions.
// Copyright (C) 2000  Kevin B. McCarty
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


//
// stararray.h - The Star Array class.  Not much more than a container class
//  for stars, with the ability to read and filter a text data file on the fly.
//

#ifndef _STAR_ARRAY_H_
#define _STAR_ARRAY_H_

#include <fstream>
#include <vector>
#include <string>
#include <cstdlib>
#include <cstdio>
using namespace std;

#include "star.h"

// A typedef to specify the kind of change made to ArrayRules when
//  setting a new set of rules:
typedef enum { FILE_CHANGE, LOCATION_CHANGE, RADIUS_CHANGE,
	       COORDINATE_CHANGE, FILTER_CHANGE, ORIENTATION_CHANGE,
	       DECORATIONS_CHANGE, NO_CHANGE } Changetype;

// Specify the character used to separate records in the data files
const char RECORD_DELIMITER = '$';

// and the maximum length of each record (note: characters beyond this
//  length will not cause any problems, they will just be ignored)
const unsigned int RECORD_MAX_LENGTH = 2000;

// and the number of stars on the display to label, if the "Landmark Stars"
//  option is checked
const unsigned int DISPLAY_N_STARS = 8;

// Do not load more than this number of stars into memory at once, to
//  avoid DOS'ing oneself.  Note that 1000 stars ~ 1 megabyte RAM
const unsigned int MAX_STARS = 20000;


class StarArray {
 private:
  vector<Star> Array;
  unsigned int TotalStars;
  Rules ArrayRules;

  StarArray(const StarArray &rhs);                 // these should never
  StarArray & operator= (const StarArray &rhs);    // be implemented.

  void toCelestial();  // convert coordinates for all the stars,
  void toGalactic();   //  as specified.  Not public because they should
                       //  only be controlled by StarArray::SetRules().

  void Read(ifstream &);   // reads and filters data file
  void Sort();             // sorts the array by x-position in local coords

 public:
  StarArray() : Array() { TotalStars = 0; }
  // no other constructors needed, memory management is automatic in STL

  inline unsigned int size() const { return Array.size(); }
  inline unsigned int totalstars() const { return TotalStars; }
  inline Star & operator [] (unsigned int i) { return Array[i]; }

  // look up stars containing a certain substring in their names and create
  //  a StarArray of them (ignores the Rules member)
  void Search(const char *searchstring, StringList filelist,
	      Rules rules, bool casesensitive = false,
	      bool exactmatch /* i.e. not a substring */ = false,
	      bool exitonmatch /* i.e. return only one match */ = false);

  // set rules for filtering and update the Array of stars using those rules
  //  (default: if type of change is not supplied, assume the most radical 
  //  change)
  bool SetRules(Rules, Changetype ruleschange = FILE_CHANGE);

  // displays all the stars, plus a grid and some other decorations
  void Display(StarViewer *sv) const;

  // plots the stars on an HR diagram in range dimmag - brightmag
  void Diagram(StarViewer *sv, double brightmag, double dimmag) const; 
};

#endif
