// StarPlot - A program for interactively viewing 3D maps of stellar positions.
// Copyright (C) 2000  Kevin B. McCarty
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 
/*
  menuops.cc
  This file contains code to set up the StarPlot menu bar and to deal with
  the toggle and radio button options in the StarPlot Options menu.
*/

#include <gtk/gtk.h>
#include "starplot.h"
#include "xpmdata.h"
#include "../version.h"

/* Here are all the items in the StarPlot menu. 
 *  Unimplemented items have been commented out.
 *  Note: C++ compilers don't like having non-void callback functions in the
 *  GtkItemFactoryEntry struct.  It interferes with strict type checking.
 */

GtkItemFactoryEntry menu_items[] = {        /*       fn.        */
  /* menu path             acc. key    callback fn. param  type */
  { "/_File",                   0,      0,            0, "<Branch>" },
  { "/File/_Open Star Database...",  0, file_open,    0, "" },
  { "/File/_Merge Star Database...", 0, file_merge,   0, "" },
//  { "/File/_Open Parameter File...", 0, params_open,  0, "" },
//  { "/File/_Save Chart as PNG...",   0, file_save,    0, "" },
//  { "/File/Save Chart _Parameters...",0,file_params,  0, "" },
  { "/File/-----",              0,     0,             0, "<Separator>" },
  { "/File/_Quit",              0,     gtk_main_quit, 0, "" },

  { "/_Chart",                  0,     0,             0, "<Branch>" },
  { "/Chart/_Define Chart...",  0,     chart_define,  0, "" },
  { "/Chart/_Orientation...",   0,     chart_orient,  0, "" },
  { "/Chart/Star _Filter...",   0,     chart_filter,  0, "" },

  { "/_Options",                0,     0,             0, "<Branch>" },
  { "/Options/Toggle _Bars",   "<control>B", toggle_bars,   0, "<CheckItem>" },
  { "/Options/Toggle G_rid",   "<control>R", toggle_grid,   0, "<CheckItem>" },
  { "/Options/Toggle _Legend", "<control>L", toggle_legend, 0, "<CheckItem>" },
  { "/Options/-----",           0,            0,      0, "<Separator>" },

  // star label and star diameter submenus
  MENU_DRAWING_OPTIONS, // #defined in starplot.h

  // coordinates submenu
  { "/Options/_Coordinate System",  0,      0,          0, "<Branch>" },
  { "/Options/Coordinate System/_Celestial", 
    "<control>C", cel_coords,       0, "<RadioItem>" },
  { "/Options/Coordinate System/_Galactic",
    "<control>G", gal_coords,       0, "/Options/Coordinate System/Celestial"},

  { "/_Stars",                  0,          0,          0, "<Branch>" },
  { "/Stars/Chart _Data...",  "<control>D", starinfo,   0, "" },
  { "/Stars/_HR Diagram...",  "<control>H", hrdiagram,  0, "" },
  { "/Stars/_Search...",      "<control>S", starsearch, 0, "" },

  { "/_Help",                   0,          0,          0, "<LastBranch>" },
  { "/Help/_About...",          0,          abouthelp,  0, "" },
//  { "/Help/_View Files",      0,            0,        0, "" },
};

const char *radio_item_names[MENU_DRAWING_OPTIONS_SIZE] = {
  "/Options/Star Labels/Landmark Stars", 
  "/Options/Star Labels/All Star Names",
  "/Options/Star Labels/Numerical Index",
  "/Options/Star Labels/Labels Off",
  "/Options/Star Diameters/By Magnitude",
  "/Options/Star Diameters/By MK Class"
};

/* my_gtk_main_menu(): This function prepares the StarPlot menu bar. */

GtkWidget *optionmenu_items[2][MENU_DRAWING_OPTIONS_SIZE];

void my_gtk_main_menu (GtkWidget *window, GtkWidget **menubar)
{
  GtkWidget *item;
  GtkItemFactory *item_factory;
  GtkAccelGroup *accel_group;
  gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);
  
  accel_group = gtk_accel_group_new ();
       
  /* This function initializes the item factory.
     Param 1: The type of menu - can be GTK_TYPE_MENU_BAR, GTK_TYPE_MENU,
     or GTK_TYPE_OPTION_MENU.
     Param 2: The path of the menu.
     Param 3: A pointer to a gtk_accel_group.  The item factory sets up
     the accelerator table while generating menus.
  */

  item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
				       accel_group);

  /* This function generates the menu items. Pass the item factory,
     the number of items in the array, the array itself, and any
     callback data for the the menu items. */
  gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);

  /* Attach the new accelerator group to the window. */
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

  /* The following code sets the check buttons in the Options menu to be
     toggled ON by default, matching the initial state of the program.

     According to the GTK documentation, I should not use the "active" field
     of GtkCheckMenuItem directly, so below I should replace each line
       (GTK_CHECK_MENU_ITEM (item))->active = TRUE;
     with the line
       gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), TRUE);
     However, when I do this, StarPlot segfaults.  Fixes welcomed.
  */

  item = gtk_item_factory_get_widget (item_factory, "/Options/Toggle Bars");
  (GTK_CHECK_MENU_ITEM (item))->active = TRUE;
  item = gtk_item_factory_get_widget (item_factory, "/Options/Toggle Grid");
  (GTK_CHECK_MENU_ITEM (item))->active = TRUE;
  item = gtk_item_factory_get_widget (item_factory, "/Options/Toggle Legend");
  (GTK_CHECK_MENU_ITEM (item))->active = TRUE;

  /* Get the radio button widgets which are in the Options menu.  We need
   *  to have them available globally in order to make them consistent between
   *  the main and HR diagram window menus.
   */
  for (unsigned int i = 0; i < MENU_DRAWING_OPTIONS_SIZE; i++) {
    item = gtk_item_factory_get_widget (item_factory, radio_item_names[i]);
    optionmenu_items[0][i] = item;
  }

  if (menubar)
    /* Finally, return the actual menu bar created by the item factory. */ 
    *menubar = gtk_item_factory_get_widget (item_factory, "<main>");
}


/* my_gtk_button_bar(): This function creates a button bar with any
   combination of OK, Defaults and Cancel buttons, and packs it into
   a vbox or hbox given as an argument.  It will NOT, however, set
   up the callback functions.  Use NULL for any of the first
   three arguments for which a button should not be created.
*/
void my_gtk_button_bar(GtkWidget **OK, GtkWidget **defaults,
		       GtkWidget **cancel, GtkWidget *insertionbox)
{
  GtkWidget *buttonbox = gtk_hbox_new (TRUE, 0);
  gtk_widget_show (buttonbox);
  gtk_box_pack_start (GTK_BOX (insertionbox), buttonbox, FALSE, FALSE, 0);
 
  if (OK) {
    *OK = gtk_button_new_with_label ("OK");
    gtk_widget_set_usize (*OK, BUTTONWIDTH, BUTTONHEIGHT);
    gtk_box_pack_start (GTK_BOX (buttonbox), *OK, TRUE, FALSE, 0);
    gtk_widget_show (*OK);
  }

  if (defaults) {
    *defaults = gtk_button_new_with_label ("Defaults");
    gtk_widget_set_usize (*defaults, BUTTONWIDTH, BUTTONHEIGHT);
    gtk_box_pack_start (GTK_BOX (buttonbox), *defaults, TRUE, FALSE, 0);
    gtk_widget_show (*defaults);
  }

  if (cancel) {
    *cancel = gtk_button_new_with_label ("Cancel");
    gtk_widget_set_usize (*cancel, BUTTONWIDTH, BUTTONHEIGHT);
    gtk_box_pack_start (GTK_BOX (buttonbox), *cancel, TRUE, FALSE, 0);
    gtk_widget_show (*cancel);
  }

  return;
}

/* my_gtk_buttons_connect_destroy(): Connects all the buttons supplied
   as arguments to the gtk_destroy_widget() function.  Should be called
   after any other signal connections.
*/
void my_gtk_buttons_connect_destroy(GtkWidget *OK, GtkWidget *defaults,
				    GtkWidget *cancel, GtkWidget *window)
{
  if (OK)
    gtk_signal_connect_object (GTK_OBJECT (OK), "clicked", 
			       GTK_SIGNAL_FUNC (gtk_widget_destroy),
			       GTK_OBJECT (window));
  if (defaults)
    gtk_signal_connect_object (GTK_OBJECT (defaults), "clicked", 
			       GTK_SIGNAL_FUNC (gtk_widget_destroy),
			       GTK_OBJECT (window));
  if (cancel)
    gtk_signal_connect_object (GTK_OBJECT (cancel), "clicked", 
			       GTK_SIGNAL_FUNC (gtk_widget_destroy),
			       GTK_OBJECT (window));
  return;
}


/* The following trivial functions each carry out one of the options in
   the Options menu.
*/

void toggle_bars() 
{ starrules.StarBars = !starrules.StarBars; redraw_all(DECORATIONS_CHANGE); }

void toggle_grid()
{ starrules.ChartGrid = !starrules.ChartGrid; redraw_all(DECORATIONS_CHANGE); }

void toggle_legend()
{
  starrules.ChartLegend = !starrules.ChartLegend;
  redraw_all(DECORATIONS_CHANGE); 
}

void labels_all()
{ 
  static bool enabled = true;

  // need all these tests to ensure that (a) this function doesn't cause
  //  itself to be called again by playing with radio button toggle states;
  //  (b) the current setting is actually _being changed_; (c) this is not
  //  being called as a result of a _different_ radio button being toggled
  //  ON, thus toggling this one OFF.

  if (enabled) {
    enabled = false;
    {
      starrules.StarLabels = STRING_LABEL;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][1]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][1]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void labels_landmk()
{ 
  static bool enabled = true;

  if (enabled) {
    enabled = false;
    {
      starrules.StarLabels = LANDMARK_LABEL;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][0]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][0]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void labels_numerical()
{
  static bool enabled = true;

  if (enabled) {
    enabled = false;
    {
      starrules.StarLabels = NUMBER_LABEL;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][2]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][2]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void labels_off()
{
  static bool enabled = true;

  if (enabled) {
    enabled = false;
    {
      starrules.StarLabels = NO_LABEL;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][3]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][3]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void mk_diameters()
{
  static bool enabled = true;

  if (enabled) {
    enabled = false;
    {
      starrules.StarDiameters = MK_DIAMETERS;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][5]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][5]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void mag_diameters()
{
  static bool enabled = true;

  if (enabled) {
    enabled = false;
    {
      starrules.StarDiameters = MAGNITUDE_DIAMETERS;
      gtk_check_menu_item_set_active 
	(GTK_CHECK_MENU_ITEM (optionmenu_items[0][4]), TRUE);
      if (hr_viewer)
	gtk_check_menu_item_set_active 
	  (GTK_CHECK_MENU_ITEM (optionmenu_items[1][4]), TRUE);
      redraw_all(DECORATIONS_CHANGE); 
    }
    enabled = true;
  }
}

void cel_coords()
{
  if (! starrules.CelestialCoords) {
    starrules.CelestialCoords = true;
    starrules.ChartLocation = starrules.ChartLocation.toCelestial();
    redraw_all(COORDINATE_CHANGE);
  }
  return;
}

void gal_coords()
{ 
  if (starrules.CelestialCoords) {
    starrules.CelestialCoords = false;
    starrules.ChartLocation = starrules.ChartLocation.toGalactic();
    redraw_all(COORDINATE_CHANGE);
  }
  return;
}


/* These callbacks are for the button bar immediately underneath the menu. */

void button_zoom(GtkWidget *w, gpointer zoomfactor)
{
  double zf = *(double *)zoomfactor;

  // zoomfactor is > 1.0 to zoom in, < 1.0 to zoom out
  starrules.ChartRadius /= zf;
  if (starrules.ChartAutosetDimmestMagnitude && zf != 1.0)
    starrules.ChartDimmestMagnitude = automagnitude(starrules.ChartRadius);
  redraw_all(RADIUS_CHANGE);
  return;
}

void button_rotate(GtkWidget *w, gpointer angle)
{
  double a = *(double *)angle;
  starrules.ChartOrientation
    = SolidAngle(starrules.ChartOrientation.getPhi() + a,
		 starrules.ChartOrientation.getTheta());
  redraw_all(ORIENTATION_CHANGE);
  return;
}

void button_tilt(GtkWidget *w, gpointer angle)
{
  double a = *(double *)angle;
  starrules.ChartOrientation
    = SolidAngle(starrules.ChartOrientation.getPhi(),
		 starrules.ChartOrientation.getTheta() + a);
  redraw_all(ORIENTATION_CHANGE);
  return;
}

void button_magchange(GtkWidget *w, gpointer magchange)
{
  double m = *(double *)magchange;
  starrules.ChartDimmestMagnitude += m;
  redraw_all(FILTER_CHANGE);
  return;
}


/* Define the button bar: */

void my_gtk_push_buttons (GtkWidget *window, GtkWidget **buttonbar)
{
  GtkWidget *vbox, *buttons[9], *pixmapw;
  GtkTooltips *tooltips;
  GtkStyle *style = gtk_widget_get_style (window);
  GtkStyle *blackbackground = gtk_style_copy (style);
  GdkPixmap *pixmap, *mask;

  static double zoominfactor = 2.0, zoomoutfactor = 1.0 / zoominfactor;
  static double leftangle = 0.5 HOURS, rightangle = -leftangle;
  static double northangle = 5.0 DEGREES, southangle = -northangle;
  static double magup = -0.5, magdown = -magup;

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);

  *buttonbar = gtk_event_box_new();
  gtk_container_add (GTK_CONTAINER (*buttonbar), vbox);
  gtk_widget_show (*buttonbar);

  blackbackground->bg[GTK_STATE_NORMAL] = blackbackground->black;
  gtk_widget_set_style (*buttonbar, blackbackground);

  for (unsigned int i = 0; i < 9; i++) {
    // this creates the button icons; they are defined in the static
    //  array `xpmdata' in the xpmdata.h header file.
    pixmap = gdk_pixmap_colormap_create_from_xpm_d 
      (0, gtk_widget_get_colormap(window), &mask,
       &style->bg[GTK_STATE_NORMAL], (gchar **)xpmdata[i]);
    pixmapw = gtk_pixmap_new (pixmap, mask);
    gdk_pixmap_unref(pixmap);
    gdk_pixmap_unref(mask);

    // create the buttons themselves
    gtk_widget_show (pixmapw);
    buttons[i] = gtk_button_new();
    gtk_container_add (GTK_CONTAINER(buttons[i]), pixmapw);
    gtk_widget_show (buttons[i]);
    gtk_box_pack_start (GTK_BOX (vbox), buttons[i], FALSE, TRUE, 0);
  }

  // connect the callback functions to the buttons
  gtk_signal_connect (GTK_OBJECT (buttons[0]), "clicked",
		      GTK_SIGNAL_FUNC (file_open), 0);
  gtk_signal_connect (GTK_OBJECT (buttons[1]), "clicked",
		      GTK_SIGNAL_FUNC (button_zoom), &zoominfactor);
  gtk_signal_connect (GTK_OBJECT (buttons[2]), "clicked",
		      GTK_SIGNAL_FUNC (button_zoom), &zoomoutfactor);
  gtk_signal_connect (GTK_OBJECT (buttons[3]), "clicked",
		      GTK_SIGNAL_FUNC (button_rotate), &leftangle);
  gtk_signal_connect (GTK_OBJECT (buttons[4]), "clicked",
		      GTK_SIGNAL_FUNC (button_rotate), &rightangle);
  gtk_signal_connect (GTK_OBJECT (buttons[5]), "clicked",
		      GTK_SIGNAL_FUNC (button_tilt), &northangle);
  gtk_signal_connect (GTK_OBJECT (buttons[6]), "clicked",
		      GTK_SIGNAL_FUNC (button_tilt), &southangle);
  gtk_signal_connect (GTK_OBJECT (buttons[7]), "clicked",
		      GTK_SIGNAL_FUNC (button_magchange), &magup);
  gtk_signal_connect (GTK_OBJECT (buttons[8]), "clicked",
		      GTK_SIGNAL_FUNC (button_magchange), &magdown);

  // and set some popup help tips
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, buttons[0], "Open star database...", 0);
  gtk_tooltips_set_tip (tooltips, buttons[1], "Zoom in", 0);
  gtk_tooltips_set_tip (tooltips, buttons[2], "Zoom out", 0);
  gtk_tooltips_set_tip (tooltips, buttons[3],
			"Rotate chart clockwise about its axis", 0);
  gtk_tooltips_set_tip (tooltips, buttons[4], 
			"Rotate chart counterclockwise about its axis", 0);
  gtk_tooltips_set_tip (tooltips, buttons[5], 
			"Tilt chart north pole towards you", 0);
  gtk_tooltips_set_tip (tooltips, buttons[6], 
			"Tilt chart south pole towards you", 0);
  gtk_tooltips_set_tip (tooltips, buttons[7], 
			"Decrease magnitude limit (Show fewer stars)", 0);
  gtk_tooltips_set_tip (tooltips, buttons[8], 
			"Increase magnitude limit (Show more stars)", 0);
  return;
}


/* Silly little "about" dialog */
void abouthelp()
{
  GtkWidget *popup, *OK, *topbox, *label[7];
  
  popup = gtk_dialog_new ();
  gtk_window_set_policy (GTK_WINDOW (popup), FALSE, FALSE, TRUE);
  gtk_window_set_title (GTK_WINDOW (popup), "About StarPlot");
  
  topbox = gtk_vbox_new (TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (topbox), 10);
  gtk_widget_show (topbox);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (popup)->vbox), topbox,
		      TRUE, TRUE, 0);
  
  label[0] = gtk_label_new ("StarPlot version " STARPLOT_VERSION);
  label[1] = gtk_label_new ("");
  // I hope 0xA9 is the copyright symbol in most char sets
  label[2] = gtk_label_new ("Copyright \xA9 2000, 2001 Kevin B. McCarty");
  label[3] = gtk_label_new ("under the GNU General Public License");
  label[4] = gtk_label_new ("");
  label[5] = gtk_label_new ("For help, please see the HTML documentation");
  label[6] = gtk_label_new ("in " DOCDIR "/html");

  for (unsigned int i = 0; i < 7; i++) {
    gtk_box_pack_start (GTK_BOX (topbox), label[i], TRUE, TRUE, 0);
    gtk_widget_show (label[i]);
  }

  my_gtk_button_bar (&OK, NULL, NULL, GTK_DIALOG (popup)->action_area);
  my_gtk_buttons_connect_destroy (OK, NULL, NULL, popup);
  
  gtk_window_set_focus (GTK_WINDOW (popup), OK);
  gtk_widget_show (popup);
}

