// StarPlot - A program for interactively viewing 3D maps of stellar positions.
// Copyright (C) 2000  Kevin B. McCarty
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#ifndef _STARPLOT_H
#define _STARPLOT_H

#include <gtk/gtk.h>
#include <cmath>
#include <cstdio>
using namespace std;

#include "../classes/stararray.h"
#include "gtkviewer.h"

#ifndef DATADIR // it should be defined in this directory's Makefile
#define DATADIR "/usr/local/share/starplot"
#endif

#ifndef DOCDIR // likewise
#define DOCDIR "/usr/local/share/doc/starplot"
#endif

#define DEGREES  * M_PI / 180
#define HOURS    * M_PI / 12

const unsigned int STARPLOT_WIDTH  = 550;
const unsigned int STARPLOT_HEIGHT = 500;

const StringList DEFAULT_FILENAMES =
	StringList(DATADIR "/sample.stars"); // DATADIR is defined in Makefile
const Vector     DEFAULT_CENTER      = Vector(0,0,0);
const SolidAngle DEFAULT_ORIENTATION = SolidAngle(21 HOURS, +30 DEGREES);
const double     DEFAULT_RADIUS      = 10 /* light-years */;
const float      DEFAULT_DIM_MAG     = +25;
const float      DEFAULT_BRIGHT_MAG  = -25;
const bool       DEFAULT_AUTOSET_MAG = true; /* automatically set dim magnitude
						limit as a fn. of radius? */
const bool       DEFAULT_MODE        = true; /* celestial coordinates */

// If you edit the following, you also must reorder the menu items in
//  menuops.cc so that this default is the first of the starlabel menu
//  items.  Otherwise the initial toggle setting will not match this default.
const Diametertype DEFAULT_DIAMETERS = MAGNITUDE_DIAMETERS;
const Labeltype  DEFAULT_SHOW_LABELS = LANDMARK_LABEL;

const bool       DEFAULT_SHOW_BARS   = true;
const bool       DEFAULT_SHOW_GRID   = true;
const bool       DEFAULT_SHOW_LEGEND = true;

const Rules initrules = { DEFAULT_FILENAMES,  
			  DEFAULT_CENTER,
			  DEFAULT_ORIENTATION,
			  DEFAULT_RADIUS, 
			  DEFAULT_DIM_MAG, 
			  DEFAULT_BRIGHT_MAG,
			  DEFAULT_AUTOSET_MAG,
			  /* default star filter: show all star classes */
			  { true, true, true, true, true, 
			    true, true, true, true, true },
			  DEFAULT_MODE, 
			  DEFAULT_DIAMETERS,
			  DEFAULT_SHOW_LABELS,
			  DEFAULT_SHOW_BARS, 
			  DEFAULT_SHOW_GRID,
			  DEFAULT_SHOW_LEGEND
};

const unsigned int BUTTONWIDTH = 75;  // global default value for size
const unsigned int BUTTONHEIGHT = 22; //  of OK / Cancel buttons

const unsigned int HR_WIDTH  = 480;
const unsigned int HR_HEIGHT = 360;
const double HR_BRIGHT_MAG = -8;
const double HR_DIM_MAG    = +12;


// these are options shared between the main window and HR diagram window
//  menus, #defined here to ensure consistency

#define MENU_DRAWING_OPTIONS_SIZE 6
extern const char *radio_item_names[MENU_DRAWING_OPTIONS_SIZE];

#define MENU_DRAWING_OPTIONS                                                  \
                                                                              \
  /* star label submenu */                                                    \
{ "/Options/_Star Labels", 0, 0,   0, "<Branch>" },                           \
{ "/Options/Star Labels/Landmar_k Stars",                                     \
    "<control>K", labels_landmk,   0, "<RadioItem>" },                        \
{ "/Options/Star Labels/All Star _Names",                                     \
    "<control>N", labels_all,      0, "/Options/Star Labels/Landmark Stars" },\
{ "/Options/Star Labels/Numerical _Index",                                    \
    "<control>I", labels_numerical,0, "/Options/Star Labels/Landmark Stars" },\
{ "/Options/Star Labels/Labels _Off",                                         \
    "<control>O", labels_off,      0, "/Options/Star Labels/Landmark Stars" },\
                                                                              \
  /* star diameter submenu */                                                 \
{ "/Options/Star _Diameters", 0, 0, 0, "<Branch>" },                          \
{ "/Options/Star Diameters/By _Magnitude",                                    \
    0, mag_diameters, 0, "<RadioItem>" },                                     \
{ "/Options/Star Diameters/By M_K Class",                                     \
    0, mk_diameters,  0, "/Options/Star Diameters/By Magnitude" }


// Menu item callbacks, defined elsewhere in this directory:

// defined in filedialogs.cc
extern void file_open();
extern void file_merge();
extern void params_open();
extern void file_save();
extern void file_params();

// in chartdialogs.cc
extern void chart_define();
extern void chart_orient();
extern void chart_filter();

// in menuops.cc
extern void toggle_bars();
extern void toggle_grid();
extern void toggle_legend();
extern void labels_all();
extern void labels_landmk();
extern void labels_off();
extern void labels_numerical();
extern void mk_diameters();
extern void mag_diameters();
extern void cel_coords();
extern void gal_coords();
extern void abouthelp();

// in infodialogs.cc
extern void starinfo();
extern void starsearch();

// in hrdiagram.cc
extern void hrdiagram();

// Button bar callbacks, defined in menuops.cc
extern void button_zoom(GtkWidget *w, gpointer zoomfactor);
extern void button_rotate(GtkWidget *w, gpointer angle);
extern void button_tilt(GtkWidget *w, gpointer angle);
extern void button_magchange(GtkWidget *w, gpointer magchange);

// Useful adjunct functions
extern double automagnitude (double radius);
extern void my_gtk_star_popup (Star &);
extern void my_gtk_main_menu (GtkWidget *window, GtkWidget **menubar);
extern void my_gtk_push_buttons (GtkWidget *window, GtkWidget **buttonbar);
extern void my_gtk_button_bar (GtkWidget **OKbtn, GtkWidget **Defaultsbtn,
			       GtkWidget **Cancelbtn, GtkWidget *box);
extern void my_gtk_buttons_connect_destroy (GtkWidget *btn1, GtkWidget *btn2,
					    GtkWidget *btn3, GtkWidget *win);
extern void hr_redraw_display(GtkWidget *hr_display);
extern void update_info(GtkWidget *info_clist);
extern void redraw_all (Changetype);

// for synching HR and main window menu radio buttons:
extern GtkWidget *optionmenu_items[2][MENU_DRAWING_OPTIONS_SIZE]; 

extern Rules starrules;              // By default set equal to initrules
extern StarArray *stararray;         // The StarArray used by StarPlot
extern GTKViewer *viewer;            // The GTKViewer used by StarPlot
extern GtkWidget *drawing_area;      // The GTK_DRAWING_AREA used
extern GtkWidget *statusline;        // status text at bottom of window

extern GTKViewer *hr_viewer;         // HR Diagram GTKViewer
extern GtkWidget *hr_drawing_area;   // HR Diagram GTK_DRAWING_AREA
extern GtkWidget *info_clist;        // Chart Data list of stars

#endif // #include _STARPLOT_H
